using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml;
using System.Text.RegularExpressions;
using Au.Gov.Sbr.Xbrl.Document;
using DataContracts;
using Au.Gov.Sbr.Xbrl.Metadata;
using System.Runtime.CompilerServices;

namespace Ato.EN.IntegrationServices.CodeGenerationRDTIS
{
    public class RDTIS2018Consumer
    {

        private static Regex StripWhitespace = new Regex(@"\s*");

        #region Helper/Validation Methods
        private static bool? ToBoolean(string str)
        {
            bool returnValue;
            string value = (str ?? "").Trim();
            if (value == "0" || value == "1")
                return Convert.ToBoolean(Convert.ToInt32(value));
            
            if (Boolean.TryParse(value, out returnValue))
                return returnValue;
            return null;
        }

        private static string GetValueOrEmpty(bool? val)
        {
            return (val.HasValue) ? val.ToString() : string.Empty;
        }

        private static string GetValueOrEmpty(DateTime? val)
        {
            return (val.HasValue) ? val.ToString() : string.Empty;
        }

        /// <summary>
        /// Derives the SBR element location.
        /// </summary>
        /// <param name="sbrElement">The SBR element.</param>
        /// <param name="addOccurrenceIndex">if set to <c>true</c> [add occurrence the occurrence index to the path].</param>
        /// <returns></returns>
        private static string DeriveSbrElementLocation(SBRElement sbrElement, bool addOccurrenceIndex = false)
        {
            if (sbrElement == null) return "/xbrli:xbrl";
            int xpathOccurrence = sbrElement.OccurrenceIndex + 1;
            string location;

            if (addOccurrenceIndex)
            {
                location = "/" + sbrElement.NamespacePrefix + ":" + sbrElement.Name + "[" +
                           xpathOccurrence + "]";
            }
            else
            {
                location = "/" + sbrElement.NamespacePrefix + ":" + sbrElement.Name;
            }

            if (sbrElement.IsInTuple)
            {
                location = DeriveSbrElementLocation(sbrElement.ParentTuple, true) + location;
            }
            else
            {
                location = "/xbrli:xbrl" + location;
            }

            return location;
        }

        private static bool IsMatch(int? field, string expression, RegexOptions options = RegexOptions.None)
        {
            if (field == null)
                return false;
            else
                return Regex.IsMatch(Convert.ToString(field.Value), expression, options);
        }

        private static bool IsMatch(string field, string expression, RegexOptions options = RegexOptions.None)
        {
            if (field == null)
                return false;
            else
                return Regex.IsMatch(field, expression, options);
        }

        public DateTime AsDate(string dateAsString)
        {
            DateTime response = DateTime.MinValue;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date;
            }

            return response;
        }

         #endregion

        public RDTIS2018 Consume(SBRDocument sbrDocument)
        {
            List<ProcessMessageDocument> errors;
            return Consume(sbrDocument, false, out errors);
        }

        /// <summary>
        /// Allows for validation of rules that can not be generated and must run at a lower level.
        /// A sub-class will need to be created with an override for this method - which does nothing in its base implementation.
        /// </summary>
        /// <param name="sbrDocument">The SBR document.</param>
        /// <param name="populatedReportElementsMap">The populated report elements map can be used to find data elements and tuples.</param>
        /// <param name="contextArray">The context array can be used to find context values.</param>
        /// <param name="errors">The errors collection is to be populated with validation results.</param>
        protected virtual void ValidateUnsupportedRules(SBRDocument sbrDocument, IDictionary<string, SBRElement> populatedReportElementsMap, SBRContext[] contextArray, List<ProcessMessageDocument> errors)
        {

        }

        public RDTIS2018 Consume(SBRDocument sbrDocument, bool validationMode, out List<ProcessMessageDocument> errors)
        {
            RDTIS2018 report;
            report = new RDTIS2018();
            errors = new List<ProcessMessageDocument>();

            SBRContext[] contextArray;
            IDictionary<string, SBRElement> populatedReportElementsMap;

            contextArray = sbrDocument.ContextsMap.Values.ToArray();
            populatedReportElementsMap = sbrDocument.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

            if (validationMode) ValidateUnsupportedRules(sbrDocument, populatedReportElementsMap, contextArray, errors);

            /// <summary>
            /// Current element being extracted
            /// </summary>
            SBRElement currentValue = null;
            /// <summary>
            /// Current elements being extracted
            /// </summary>
            IList<SBRElement> currentValues = null;
            /// <summary>
            /// Current occurrances being extracted
            /// </summary>
            IList<SBRElement> currentOccurrences = null;
            /// <summary>
            /// Working Variable for date time variables
            /// </summary>
            DateTime currentDateTimeValue;
            /// <summary>
            /// Working Variable for boolean variables
            /// </summary>
            bool currentBooleanValue;
            /// <summary>
            /// Working Variable for decimal variables
            /// </summary>
            decimal currentDecimalValue;
            /// <summary>
            /// Working Variable for integer variables
            /// </summary>
            int currentIntValue;
            /// <summary>
            /// Working Variable for long variables
            /// </summary>
            long currentLongValue;
            XmlDocumentFragment currentSegment;
            XmlNamespaceManager currentNamespaceManager;
            XmlNode currentNode;
    
            #region sbrRPContext
            IEnumerable<SBRContext> sbrRPContexts = null;
            SBRContext sbrRPContext = null;
    
            sbrRPContexts = contextArray.Where(
                c => c.Period != null && c.Period.PeriodType == SBRPeriod_PeriodType.DURATION &&
                c.Segment != null && c.Segment.ExplicitOrTypedDimensionDomains.Count == 1 && 
                (c.Segment.ExplicitOrTypedDimensionDomains.Any(dd => dd.DimensionType == SBRExplicitOrTypedDimensionDomain_Type.EXPLICIT && dd.Dimension.EndsWith("}ReportPartyTypeDimension") && ((SBRExplicitDimensionDomain)dd).Domain.EndsWith("}ReportingParty")) ));
            Dictionary<string, IList<SBRElement>> sbrPopulatedRPElementsMap = new Dictionary<string, IList<SBRElement>>();
            if (sbrRPContexts != null && sbrRPContexts.Count() > 0)
            {
                sbrRPContext = sbrRPContexts.First();
        
                report.RPId = sbrRPContext.Id;
                report.RPIdentifier = sbrRPContext.EntityIdentifier;
                report.RPIdentifierScheme = sbrRPContext.EntityScheme;
        
                report.RPCount = sbrRPContexts.Count();
                report.RPExists = true;
                report.RPOccurrenceIndex = Array.IndexOf(contextArray, sbrRPContext) + 1;
                report.RPLastOccurrenceIndex = Array.IndexOf(contextArray, sbrRPContexts.Last()) + 1;
        

                if (report.RPIdentifierScheme == "http://www.ato.gov.au/tfn") report.RPIdentifierTFN = report.RPIdentifier;
    
                report.RPStartDate = ((SBRDurationPeriod)sbrRPContext.Period).DurationStartAsDateObject;
                report.RPEndDate = ((SBRDurationPeriod)sbrRPContext.Period).DurationEndAsDateObject;
        
                sbrPopulatedRPElementsMap = new System.Collections.Generic.Dictionary<string, IList<SBRElement>>();
                for (int i = 0; i < sbrRPContexts.Count(); i++)
                {
                    sbrRPContext = sbrRPContexts.ElementAt(i);
            
                    IList<SBRElement> elements = sbrDocument.ElementsByContextId[sbrRPContext.Id];
                    if (elements != null)
                    {
                        for (int e = 0; e < elements.Count(); e++)
                        {
                            SBRElement currentElement = elements[e];
                            if (currentElement == null || currentElement.IsInTuple) continue; // Don't add tuple elements here, they will be found through a different map.
                            if (sbrPopulatedRPElementsMap.ContainsKey("{" + currentElement.Namespace + "}" + currentElement.Name))
                            {
                                sbrPopulatedRPElementsMap["{" + currentElement.Namespace + "}" + currentElement.Name].Add(currentElement);
                            }
                            else
                            {
                                sbrPopulatedRPElementsMap.Add("{" + currentElement.Namespace + "}" + currentElement.Name, new List<SBRElement> { currentElement } );
                            }
                        }
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Report.Amendment.Indicator", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS1 = ToBoolean(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Identifiers.InnovationRegistrationNumber.Identifier", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS2 = currentValues[0].Value;
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.ResearchAndDevelopmentAccountingExpenditureAddBack.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS3 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.AllocatedTotal.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS24 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Income.ResearchAndDevelopment.FeedstockRevenueTotal.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS25 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.FeedstockInput.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS26 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Income.ResearchAndDevelopment.FeedstockNet.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS27 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}ResearchAndDevelopment.GovernmentGrantsAndRecoupments.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS28 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.RecoupmentRelated.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS29 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.GrantRelated.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS30 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ResearchAndDevelopment.GovernmentGrantsRecoupmentsClawback.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS31 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.AssociatesUnpaid.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS32 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.AssociatesClaimed.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS33 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.AssociatesNotClaimed.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS34 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.AssociatesCarriedForward.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS35 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ExemptEntityOwnership.Indicator", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS36 = ToBoolean(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.AggregatedTurnoverAtOrAboveThreshold.Indicator", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS37 = ToBoolean(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ResearchAndDevelopment.TaxOffsetEntityTurnover.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS38 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ResearchAndDevelopment.TaxOffsetAdditionalEntitiesTurnover.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS52 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ResearchAndDevelopment.TaxOffsetTurnoverExclusion.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS43 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ResearchAndDevelopment.TaxOffsetTurnoverTotal.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS44 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ResearchAndDevelopment.TaxOffsetNotionalRefundable.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS45 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ResearchAndDevelopment.TaxOffsetRefundable.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS46 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ResearchAndDevelopment.TaxOffsetNotionalNonRefundable.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS47 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ResearchAndDevelopment.TaxOffsetNonRefundable.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS48 = decimal.Parse(currentValues[0].Value);
                    }
                } 
            } 
            #endregion End of Context sbrRPContext
    
            #region sbrGMContext
            IEnumerable<SBRContext> sbrGMContexts = null;
            SBRContext sbrGMContext = null;
    
            sbrGMContexts = contextArray.Where(
                c => c.Period != null && c.Period.PeriodType == SBRPeriod_PeriodType.DURATION &&
                c.Segment != null && c.Segment.ExplicitOrTypedDimensionDomains.Count == 1 && 
                (c.Segment.ExplicitOrTypedDimensionDomains.Any(dd => dd.DimensionType == SBRExplicitOrTypedDimensionDomain_Type.EXPLICIT && dd.Dimension.EndsWith("}ReportPartyTypeDimension") && ((SBRExplicitDimensionDomain)dd).Domain.EndsWith("}GroupMember")) ));
            Dictionary<string, IList<SBRElement>> sbrPopulatedGMElementsMap = new Dictionary<string, IList<SBRElement>>();
            if (sbrGMContexts != null && sbrGMContexts.Count() > 0)
            {
        
                report.GMContextCollection = new List<RDTIS2018.GMContext>();
                for (int i = 0; i < sbrGMContexts.Count(); i++)
                {
                    sbrGMContext = sbrGMContexts.ElementAt(i);
                    RDTIS2018.GMContext gmContext = new RDTIS2018.GMContext();
                    report.GMContextCollection.Add(gmContext);
            
                    gmContext.Id = sbrGMContext.Id;
                    gmContext.Identifier = sbrGMContext.EntityIdentifier;
                    gmContext.IdentifierScheme = sbrGMContext.EntityScheme;
            
                    gmContext.Count = sbrGMContexts.Count();
                    gmContext.Exists = true;
                    gmContext.OccurrenceIndex = Array.IndexOf(contextArray, sbrGMContext) + 1;
                    gmContext.LastOccurrenceIndex = Array.IndexOf(contextArray, sbrGMContexts.Last()) + 1;
            
    
                    if (gmContext.IdentifierScheme == "http://www.ato.gov.au/tfn") gmContext.IdentifierTFN = gmContext.Identifier;
        
                    gmContext.StartDate = ((SBRDurationPeriod)sbrGMContext.Period).DurationStartAsDateObject;
                    gmContext.EndDate = ((SBRDurationPeriod)sbrGMContext.Period).DurationEndAsDateObject;
            
                    sbrPopulatedGMElementsMap = new System.Collections.Generic.Dictionary<string, IList<SBRElement>>();
            
                    IList<SBRElement> elements = sbrDocument.ElementsByContextId[sbrGMContext.Id];
                    if (elements != null)
                    {
                        for (int e = 0; e < elements.Count(); e++)
                        {
                            SBRElement currentElement = elements[e];
                            if (currentElement == null) continue; // Don't add tuple elements here, they will be found through a different map.
                            if (sbrPopulatedGMElementsMap.ContainsKey("{" + currentElement.Namespace + "}" + currentElement.Name))
                            {
                                sbrPopulatedGMElementsMap["{" + currentElement.Namespace + "}" + currentElement.Name].Add(currentElement);
                            }
                            else
                            {
                                sbrPopulatedGMElementsMap.Add("{" + currentElement.Namespace + "}" + currentElement.Name, new List<SBRElement> { currentElement } );
                            }
                        }
                    }
        
                    IEnumerable<SBRElement> sbrOrganisationNameDetailsCollection;
                    if (populatedReportElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}OrganisationNameDetails", out currentValue))
                        sbrOrganisationNameDetailsCollection = currentValue.Occurrences;
                    else
                        sbrOrganisationNameDetailsCollection = null;
            
                    #region sbrOrganisationNameDetailsCollection
            
                    if (sbrOrganisationNameDetailsCollection != null)
                    {
                        gmContext.OrganisationNameDetailsCollectionExists = true;
                        gmContext.OrganisationNameDetailsCollectionCount = sbrOrganisationNameDetailsCollection.Count();
                        SBRElement sbrOrganisationNameDetails;
                        sbrOrganisationNameDetails = sbrOrganisationNameDetailsCollection.ElementAt(0);
                
                        IDictionary<string, SBRElement> sbrOrganisationNameDetailsTupleElementMap;
                        sbrOrganisationNameDetailsTupleElementMap = sbrOrganisationNameDetails.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();
                
                        if (sbrPopulatedGMElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}OrganisationNameDetails.OrganisationalNameType.Code", out currentValues))
                        {
                            if (!currentValues[0].IsNil) gmContext.RDTIS39 = currentValues[0].Value;
                        }
                
                        if (sbrPopulatedGMElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}OrganisationNameDetails.Currency.Code", out currentValues))
                        {
                            if (!currentValues[0].IsNil) gmContext.RDTIS40 = currentValues[0].Value;
                        }
                
                        if (sbrPopulatedGMElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}OrganisationNameDetails.OrganisationalName.Text", out currentValues))
                        {
                            if (!currentValues[0].IsNil) gmContext.RDTIS41 = currentValues[0].Value;
                        }
                    } 
                    #endregion End of Tuple sbrOrganisationNameDetailsCollection
            
                    if (sbrPopulatedGMElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}TaxConcession.ResearchAndDevelopment.TaxOffsetEntityTurnover.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) gmContext.RDTIS42 = decimal.Parse(currentValues[0].Value);
                    }
        } 
            } 
            #endregion End of Context sbrGMContext
    
            #region sbrRPCTAUSContext
            IEnumerable<SBRContext> sbrRPCTAUSContexts = null;
            SBRContext sbrRPCTAUSContext = null;
    
            sbrRPCTAUSContexts = contextArray.Where(
                c => c.Period != null && c.Period.PeriodType == SBRPeriod_PeriodType.DURATION &&
                c.Segment != null && c.Segment.ExplicitOrTypedDimensionDomains.Count == 2 && 
                (c.Segment.ExplicitOrTypedDimensionDomains.Any(dd => dd.DimensionType == SBRExplicitOrTypedDimensionDomain_Type.EXPLICIT && dd.Dimension.EndsWith("}ReportPartyTypeDimension") && ((SBRExplicitDimensionDomain)dd).Domain.EndsWith("}ReportingParty"))  && 
                (c.Segment.ExplicitOrTypedDimensionDomains.Any(dd => dd.DimensionType == SBRExplicitOrTypedDimensionDomain_Type.EXPLICIT && dd.Dimension.EndsWith("}ConcessionTypeDimension") && ((SBRExplicitDimensionDomain)dd).Domain.EndsWith("}Australian")) )));
            Dictionary<string, IList<SBRElement>> sbrPopulatedRPCTAUSElementsMap = new Dictionary<string, IList<SBRElement>>();
            if (sbrRPCTAUSContexts != null && sbrRPCTAUSContexts.Count() > 0)
            {
                sbrRPCTAUSContext = sbrRPCTAUSContexts.First();
        
                report.RPCTAUSId = sbrRPCTAUSContext.Id;
                report.RPCTAUSIdentifier = sbrRPCTAUSContext.EntityIdentifier;
                report.RPCTAUSIdentifierScheme = sbrRPCTAUSContext.EntityScheme;
        
                report.RPCTAUSCount = sbrRPCTAUSContexts.Count();
                report.RPCTAUSExists = true;
                report.RPCTAUSOccurrenceIndex = Array.IndexOf(contextArray, sbrRPCTAUSContext) + 1;
                report.RPCTAUSLastOccurrenceIndex = Array.IndexOf(contextArray, sbrRPCTAUSContexts.Last()) + 1;
        

                if (report.RPCTAUSIdentifierScheme == "http://www.ato.gov.au/tfn") report.RPCTAUSIdentifierTFN = report.RPCTAUSIdentifier;
    
                report.RPCTAUSStartDate = ((SBRDurationPeriod)sbrRPCTAUSContext.Period).DurationStartAsDateObject;
                report.RPCTAUSEndDate = ((SBRDurationPeriod)sbrRPCTAUSContext.Period).DurationEndAsDateObject;
        
                sbrPopulatedRPCTAUSElementsMap = new System.Collections.Generic.Dictionary<string, IList<SBRElement>>();
                for (int i = 0; i < sbrRPCTAUSContexts.Count(); i++)
                {
                    sbrRPCTAUSContext = sbrRPCTAUSContexts.ElementAt(i);
            
                    IList<SBRElement> elements = sbrDocument.ElementsByContextId[sbrRPCTAUSContext.Id];
                    if (elements != null)
                    {
                        for (int e = 0; e < elements.Count(); e++)
                        {
                            SBRElement currentElement = elements[e];
                            if (currentElement == null || currentElement.IsInTuple) continue; // Don't add tuple elements here, they will be found through a different map.
                            if (sbrPopulatedRPCTAUSElementsMap.ContainsKey("{" + currentElement.Namespace + "}" + currentElement.Name))
                            {
                                sbrPopulatedRPCTAUSElementsMap["{" + currentElement.Namespace + "}" + currentElement.Name].Add(currentElement);
                            }
                            else
                            {
                                sbrPopulatedRPCTAUSElementsMap.Add("{" + currentElement.Namespace + "}" + currentElement.Name, new List<SBRElement> { currentElement } );
                            }
                        }
                    }
            
                    if (sbrPopulatedRPCTAUSElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.ContractedExpenditureResearchServiceProvider.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS4 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTAUSElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.ContractedExpenditureNonResearchServiceProvider.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS6 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTAUSElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Remuneration.WagesAndSalaries.Salary.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS8 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTAUSElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.OtherExpenditureRegisteredResearchAgency.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS10 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTAUSElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.Feedstock.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS12 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTAUSElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.AssociatesExpenditure.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS14 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTAUSElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.DeclineInValuePlantAndDepreciatingAssets.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS16 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTAUSElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.BalancingLosses.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS18 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTAUSElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.CooperativeResearchCentreContributions.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS20 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTAUSElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.AllocatedTotal.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS22 = decimal.Parse(currentValues[0].Value);
                    }
                } 
            } 
            #endregion End of Context sbrRPCTAUSContext
    
            #region sbrRPCTFORContext
            IEnumerable<SBRContext> sbrRPCTFORContexts = null;
            SBRContext sbrRPCTFORContext = null;
    
            sbrRPCTFORContexts = contextArray.Where(
                c => c.Period != null && c.Period.PeriodType == SBRPeriod_PeriodType.DURATION &&
                c.Segment != null && c.Segment.ExplicitOrTypedDimensionDomains.Count == 2 && 
                (c.Segment.ExplicitOrTypedDimensionDomains.Any(dd => dd.DimensionType == SBRExplicitOrTypedDimensionDomain_Type.EXPLICIT && dd.Dimension.EndsWith("}ReportPartyTypeDimension") && ((SBRExplicitDimensionDomain)dd).Domain.EndsWith("}ReportingParty"))  && 
                (c.Segment.ExplicitOrTypedDimensionDomains.Any(dd => dd.DimensionType == SBRExplicitOrTypedDimensionDomain_Type.EXPLICIT && dd.Dimension.EndsWith("}ConcessionTypeDimension") && ((SBRExplicitDimensionDomain)dd).Domain.EndsWith("}Foreign")) )));
            Dictionary<string, IList<SBRElement>> sbrPopulatedRPCTFORElementsMap = new Dictionary<string, IList<SBRElement>>();
            if (sbrRPCTFORContexts != null && sbrRPCTFORContexts.Count() > 0)
            {
                sbrRPCTFORContext = sbrRPCTFORContexts.First();
        
                report.RPCTFORId = sbrRPCTFORContext.Id;
                report.RPCTFORIdentifier = sbrRPCTFORContext.EntityIdentifier;
                report.RPCTFORIdentifierScheme = sbrRPCTFORContext.EntityScheme;
        
                report.RPCTFORCount = sbrRPCTFORContexts.Count();
                report.RPCTFORExists = true;
                report.RPCTFOROccurrenceIndex = Array.IndexOf(contextArray, sbrRPCTFORContext) + 1;
                report.RPCTFORLastOccurrenceIndex = Array.IndexOf(contextArray, sbrRPCTFORContexts.Last()) + 1;
        

                if (report.RPCTFORIdentifierScheme == "http://www.ato.gov.au/tfn") report.RPCTFORIdentifierTFN = report.RPCTFORIdentifier;
    
                report.RPCTFORStartDate = ((SBRDurationPeriod)sbrRPCTFORContext.Period).DurationStartAsDateObject;
                report.RPCTFOREndDate = ((SBRDurationPeriod)sbrRPCTFORContext.Period).DurationEndAsDateObject;
        
                sbrPopulatedRPCTFORElementsMap = new System.Collections.Generic.Dictionary<string, IList<SBRElement>>();
                for (int i = 0; i < sbrRPCTFORContexts.Count(); i++)
                {
                    sbrRPCTFORContext = sbrRPCTFORContexts.ElementAt(i);
            
                    IList<SBRElement> elements = sbrDocument.ElementsByContextId[sbrRPCTFORContext.Id];
                    if (elements != null)
                    {
                        for (int e = 0; e < elements.Count(); e++)
                        {
                            SBRElement currentElement = elements[e];
                            if (currentElement == null || currentElement.IsInTuple) continue; // Don't add tuple elements here, they will be found through a different map.
                            if (sbrPopulatedRPCTFORElementsMap.ContainsKey("{" + currentElement.Namespace + "}" + currentElement.Name))
                            {
                                sbrPopulatedRPCTFORElementsMap["{" + currentElement.Namespace + "}" + currentElement.Name].Add(currentElement);
                            }
                            else
                            {
                                sbrPopulatedRPCTFORElementsMap.Add("{" + currentElement.Namespace + "}" + currentElement.Name, new List<SBRElement> { currentElement } );
                            }
                        }
                    }
            
                    if (sbrPopulatedRPCTFORElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.ContractedExpenditureResearchServiceProvider.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS5 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTFORElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.ContractedExpenditureNonResearchServiceProvider.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS7 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTFORElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Remuneration.WagesAndSalaries.Salary.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS9 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTFORElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.OtherExpenditureRegisteredResearchAgency.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS11 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTFORElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.Feedstock.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS13 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTFORElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.AssociatesExpenditure.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS15 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTFORElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.DeclineInValuePlantAndDepreciatingAssets.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS17 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTFORElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.BalancingLosses.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS19 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTFORElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.CooperativeResearchCentreContributions.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS21 = decimal.Parse(currentValues[0].Value);
                    }
            
                    if (sbrPopulatedRPCTFORElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.AllocatedTotal.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS23 = decimal.Parse(currentValues[0].Value);
                    }
                } 
            } 
            #endregion End of Context sbrRPCTFORContext
    
            #region sbrRPYPreviousContext
            IEnumerable<SBRContext> sbrRPYPreviousContexts = null;
            SBRContext sbrRPYPreviousContext = null;
    
            sbrRPYPreviousContexts = contextArray.Where(
                c => c.Period != null && c.Period.PeriodType == SBRPeriod_PeriodType.DURATION &&
                c.Segment != null && c.Segment.ExplicitOrTypedDimensionDomains.Count == 2 && 
                (c.Segment.ExplicitOrTypedDimensionDomains.Any(dd => dd.DimensionType == SBRExplicitOrTypedDimensionDomain_Type.EXPLICIT && dd.Dimension.EndsWith("}ReportPartyTypeDimension") && ((SBRExplicitDimensionDomain)dd).Domain.EndsWith("}ReportingParty"))  && 
                (c.Segment.ExplicitOrTypedDimensionDomains.Any(dd => dd.DimensionType == SBRExplicitOrTypedDimensionDomain_Type.EXPLICIT && dd.Dimension.EndsWith("}RelativePeriodDurationDimension") && ((SBRExplicitDimensionDomain)dd).Domain.EndsWith("}YPrevious")) )));
            Dictionary<string, IList<SBRElement>> sbrPopulatedRPYPreviousElementsMap = new Dictionary<string, IList<SBRElement>>();
            if (sbrRPYPreviousContexts != null && sbrRPYPreviousContexts.Count() > 0)
            {
                sbrRPYPreviousContext = sbrRPYPreviousContexts.First();
        
                report.RPYPreviousId = sbrRPYPreviousContext.Id;
                report.RPYPreviousIdentifier = sbrRPYPreviousContext.EntityIdentifier;
                report.RPYPreviousIdentifierScheme = sbrRPYPreviousContext.EntityScheme;
        
                report.RPYPreviousCount = sbrRPYPreviousContexts.Count();
                report.RPYPreviousExists = true;
                report.RPYPreviousOccurrenceIndex = Array.IndexOf(contextArray, sbrRPYPreviousContext) + 1;
                report.RPYPreviousLastOccurrenceIndex = Array.IndexOf(contextArray, sbrRPYPreviousContexts.Last()) + 1;
        

                if (report.RPYPreviousIdentifierScheme == "http://www.ato.gov.au/tfn") report.RPYPreviousIdentifierTFN = report.RPYPreviousIdentifier;
    
                report.RPYPreviousStartDate = ((SBRDurationPeriod)sbrRPYPreviousContext.Period).DurationStartAsDateObject;
                report.RPYPreviousEndDate = ((SBRDurationPeriod)sbrRPYPreviousContext.Period).DurationEndAsDateObject;
        
                sbrPopulatedRPYPreviousElementsMap = new System.Collections.Generic.Dictionary<string, IList<SBRElement>>();
                for (int i = 0; i < sbrRPYPreviousContexts.Count(); i++)
                {
                    sbrRPYPreviousContext = sbrRPYPreviousContexts.ElementAt(i);
            
                    IList<SBRElement> elements = sbrDocument.ElementsByContextId[sbrRPYPreviousContext.Id];
                    if (elements != null)
                    {
                        for (int e = 0; e < elements.Count(); e++)
                        {
                            SBRElement currentElement = elements[e];
                            if (currentElement == null || currentElement.IsInTuple) continue; // Don't add tuple elements here, they will be found through a different map.
                            if (sbrPopulatedRPYPreviousElementsMap.ContainsKey("{" + currentElement.Namespace + "}" + currentElement.Name))
                            {
                                sbrPopulatedRPYPreviousElementsMap["{" + currentElement.Namespace + "}" + currentElement.Name].Add(currentElement);
                            }
                            else
                            {
                                sbrPopulatedRPYPreviousElementsMap.Add("{" + currentElement.Namespace + "}" + currentElement.Name, new List<SBRElement> { currentElement } );
                            }
                        }
                    }
            
                    if (sbrPopulatedRPYPreviousElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/rdtis}Expense.ResearchAndDevelopment.AssociatesCarriedForward.Amount", out currentValues))
                    {
                        if (!currentValues[0].IsNil) report.RDTIS51 = decimal.Parse(currentValues[0].Value);
                    }
                } 
            } 
            #endregion End of Context sbrRPYPreviousContext
            return report;

        } // Of Consume Method
    } // Of Class
} // Of Namespace

