using System.Linq;
using System.Collections.Generic;
using DataContracts;

namespace Ato.EN.IntegrationServices.CodeGenerationFTER
{
    public partial class FTER2018Validator
    {
        #region Manual Rules - Rules marked for manual coding

        #region VR.ATO.FTER.408077
        /*  VR.ATO.FTER.408077
                If the answer to the question 'was trustee a non resident for tax purposes' is 'No' (false), then the Full period indicator must not be present and a period of non residency must NOT be present

                Legacy Rule Format:
                    WHERE IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)
                    IF [FTER36] = FALSE AND ([FTER37] <> NULL OR COUNT (TUPLE(perioddetails1.xx.xx:PeriodDetails)) > 0)
                      RETURN VALIDATION MESSAGE
                    ENDIF
                    [FTER36] = FTER:RP:Residency.TaxPurposesPersonStatus.Indicator IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)
                    [FTER37] = FTER:RP:Residency.NonResidentFullPeriod.Indicator IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)

                Technical Business Rule Format:
                    ^Trustee <> NULL AND ^FTER36 = FALSE AND (^FTER37 <> NULL OR Count(^PeriodDetails) > 0)

                Data Elements:

                    RP:^FTER36 = pyde.02.25:Residency.TaxPurposesPersonStatus.Indicator

                    RP:^FTER37 = pyde.02.00:Residency.NonResidentFullPeriod.Indicator

                    ^PeriodDetails = period1.02.01:PeriodDetails

                    ^Trustee = fter.0002.lodge.req.02.00:Trustee
                */
        protected void VRATOFTER408077(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.TrusteeCollection != null)
                response.AddRange(report.TrusteeCollection
                    .Where(f => !f.FTER36.GetValueOrDefault() && (f.FTER37 != null || Count(f.Trustee_PeriodDetailsCollectionCount) > 0))
                    .Select(f =>
                        new ProcessMessageDocument
                        {
                            Code = "CMN.ATO.FTER.408077",
                            Severity = ProcessMessageSeverity.Error,
                            Description = "Non residency question must be 'yes' (true) or periods must not be present",
                            LongDescription = "If the answer to the question 'was trustee a non resident for tax purposes' is 'No' (false), then the Full period indicator must not be present and a period of non residency must NOT be present",
                            Location = $"/xbrli:xbrl/tns:Trustee{OccurrenceIndex(f.OccurrenceIndex)}/tns:Residency.TaxPurposesPersonStatus.Indicator",
                            Parameters = new ProcessMessageParameters
                            {
                                new ProcessMessageParameter {Name = "RuleIdentifier", Value = "VR.ATO.FTER.408077"},
                                new ProcessMessageParameter {Name = "FTER36", Value = GetValueOrEmpty(f.FTER36)},
                                new ProcessMessageParameter {Name = "FTER37", Value = GetValueOrEmpty(f.FTER37)}
                            }
                        }
                    ));
        }
        #endregion // VR.ATO.FTER.408077
        #region VR.ATO.FTER.408078
        /*  VR.ATO.FTER.408078
                Trustee full period indicator or period of non-residency dates must be present

                Legacy Rule Format:
                    WHERE IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)
                    IF [FTER36] = TRUE AND ( ([FTER37] = NULL OR [FTER37] = FALSE) AND COUNT (TUPLE(perioddetails1.xx.xx:PeriodDetails)) = 0)
                      RETURN VALIDATION MESSAGE
                    ENDIF
                    [FTER36] = FTER:RP:Residency.TaxPurposesPersonStatus.Indicator IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)
                    [FTER37] = FTER:RP:Residency.NonResidentFullPeriod.Indicator IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)

                Technical Business Rule Format:
                    ^Trustee <> NULL AND ^FTER36 = TRUE AND ((^FTER37 = NULL OR ^FTER37 = FALSE) AND Count(^PeriodDetails) = 0)

                Data Elements:

                    RP:^FTER37 = pyde.02.00:Residency.NonResidentFullPeriod.Indicator

                    RP:^FTER36 = pyde.02.25:Residency.TaxPurposesPersonStatus.Indicator

                    ^PeriodDetails = period1.02.01:PeriodDetails

                    ^Trustee = fter.0002.lodge.req.02.00:Trustee
                */
        protected void VRATOFTER408078(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.TrusteeCollection != null)
                response.AddRange(report.TrusteeCollection
                    .Where(f => f.FTER36.GetValueOrDefault() && !f.FTER37.GetValueOrDefault() && Count(f.Trustee_PeriodDetailsNonResidentPartPeriodRPCollection) == 0)
                    .Select(f =>
                        new ProcessMessageDocument
                        {
                            Code = "CMN.ATO.FTER.408078",
                            Severity = ProcessMessageSeverity.Error,
                            Description = "Trustee full period indicator or period of non-residency dates must be present",
                            LongDescription = "If the answer to the question 'was trustee a non resident for tax purposes is 'Yes' (true), then the Full period indicator or Dates of non residency must be present",
                            Location = $"/xbrli:xbrl/tns:Trustee{OccurrenceIndex(f.OccurrenceIndex)}/tns:Residency.NonResidentFullPeriod.Indicator",
                            Parameters = new ProcessMessageParameters
                            {
                                new ProcessMessageParameter {Name = "RuleIdentifier", Value = "VR.ATO.FTER.408078"},
                                new ProcessMessageParameter {Name = "FTER36", Value = GetValueOrEmpty(f.FTER36)},
                                new ProcessMessageParameter {Name = "FTER37", Value = GetValueOrEmpty(f.FTER37)}
                            }
                        }
                    ));
        }
        #endregion // VR.ATO.FTER.408078
        #region VR.ATO.FTER.408079
        /*  VR.ATO.FTER.408079
                Full period indicator and dates must not both be present

                Legacy Rule Format:
                    WHERE IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)
                    IF [FTER37] = TRUE AND COUNT (TUPLE(perioddetails1.xx.xx:PeriodDetails)) > 0
                      RETURN VALIDATION MESSAGE
                    ENDIF
                    [FTER37] = FTER:RP:Residency.NonResidentFullPeriod.Indicator IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)

                Technical Business Rule Format:
                    ^Trustee <> NULL AND ^FTER37 = TRUE AND Count(^PeriodDetails) > 0

                Data Elements:

                    RP:^FTER37 = pyde.02.00:Residency.NonResidentFullPeriod.Indicator

                    ^PeriodDetails = period1.02.01:PeriodDetails

                    ^Trustee = fter.0002.lodge.req.02.00:Trustee
                */
        protected void VRATOFTER408079(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.TrusteeCollection != null)
                response.AddRange(report.TrusteeCollection
                    .Where(f => f.FTER37.GetValueOrDefault() && f.Trustee_PeriodDetailsCollectionExists && f.Trustee_PeriodDetailsCollectionCount > 0)
                    .Select(f =>
                        new ProcessMessageDocument
                        {
                            Code = "CMN.ATO.FTER.408079",
                            Severity = ProcessMessageSeverity.Error,
                            Description = "Full period indicator and dates must not both be present",
                            LongDescription = "If trustee was a non-resident for the full period indicator is 'yes' (true), then specific dates specifying the time(s) the trustee was a non-resident must not be present",
                            Location = $"/xbrli:xbrl/tns:Trustee{OccurrenceIndex(f.OccurrenceIndex)}/tns:Residency.NonResidentFullPeriod.Indicator",
                            Parameters = new ProcessMessageParameters
                            {
                                new ProcessMessageParameter {Name = "RuleIdentifier", Value = "VR.ATO.FTER.408079"},
                                new ProcessMessageParameter {Name = "FTER37", Value = GetValueOrEmpty(f.FTER37)}
                            }
                        }
                    ));
        }
        #endregion // VR.ATO.FTER.408079
        #region VR.ATO.FTER.408084
        /*  VR.ATO.FTER.408084
                Trustee non-resident period must not be before the substituted accounting period

                Legacy Rule Format:
                    IF [FTER41] <> NULL AND (ANY OCCURRENCE OF [FTER38] <> NULL AND [FTER38] < [FTER41])
                       RETURN VALIDATION MESSAGE
                    ENDIF
                    [FTER41] = FTER:RP:Period.Start.Date WHERE (TUPLE ELEMENT EXPLICIT pyin.xx.xx:Period.Type.Code = "SubstitutedAccountingPeriod") IN TUPLE(perioddetails1.xx.xx:PeriodDetails)
                    [FTER38] = FTER:RP:Period.Start.Date WHERE (TUPLE ELEMENT EXPLICIT pyin.xx.xx:Period.Type.Code = "NonResidentPartPeriod") IN TUPLE(perioddetails1.xx.xx:PeriodDetails) IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)

                Technical Business Rule Format:
                    ^FTER41 <> NULL AND AnyOccurrence(^FTER38, ^FTER38 <> NULL AND ^FTER38 < ^FTER41)

                Data Elements:

                    RP:^FTER38 = pyin.02.00:Period.Start.Date WHERE [Period.Type.Code = 'NonResidentPartPeriod']

                    RP:^FTER41 = pyin.02.00:Period.Start.Date WHERE [Period.Type.Code = 'SubstitutedAccountingPeriod']
                */
        protected void VRATOFTER408084(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.FTER41 != null && report.TrusteeCollection != null)
                response.AddRange(report.TrusteeCollection
                    .Where(f => f.Trustee_PeriodDetailsNonResidentPartPeriodRPCollection?.Find(r => r.FTER38 != null && r.FTER38.GetValueOrDefault() < report.FTER41.GetValueOrDefault()) != null)
                    .Select(f =>
                        new ProcessMessageDocument
                        {
                            Code = "CMN.ATO.FTER.408084",
                            Severity = ProcessMessageSeverity.Error,
                            Description = "Trustee non-resident period must not be before the substituted accounting period",
                            LongDescription = "The start time during the relevant period at which the trustee was a non-resident for tax purposes must not be earlier than the substituted accounting period",
                            Location = "/xbrli:xbrl/tns:Trustee/tns:PeriodDetails" + OccurrenceIndex(1 + f.Trustee_PeriodDetailsNonResidentPartPeriodRPCollection.FindIndex(r => r.FTER38 != null && r.FTER38.GetValueOrDefault() < report.FTER41.GetValueOrDefault())) + "/tns:Period.Start.Date",
                            Parameters = new ProcessMessageParameters
                            {
                                new ProcessMessageParameter {Name = "RuleIdentifier", Value = "VR.ATO.FTER.408084"},
                                new ProcessMessageParameter {Name = "FTER41", Value = GetValueOrEmptyDate(report.FTER41)}
                            }
                        }
                    ));
        }
        #endregion // VR.ATO.FTER.408084
        #region VR.ATO.FTER.408085
        /*  VR.ATO.FTER.408085
                Trustee non-resident period must not be before the election commencement time

                Legacy Rule Format:
                    IF [FTER43] <> NULL AND (ANY OCCURRENCE OF [FTER38] <> NULL AND [FTER38] < [FTER43])
                       RETURN VALIDATION MESSAGE
                    ENDIF
                    [FTER43] = FTER:RP:Elections.Commencement.Date
                    [FTER38] = FTER:RP:Period.Start.Date WHERE (TUPLE ELEMENT EXPLICIT pyin.xx.xx:Period.Type.Code = "NonResidentPartPeriod") IN TUPLE(perioddetails1.xx.xx:PeriodDetails) IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)

                Technical Business Rule Format:
                    ^FTER43 <> NULL AND AnyOccurrence(^FTER38, ^FTER38 <> NULL AND ^FTER38 < ^FTER43)

                Data Elements:

                    RP:^FTER38 = pyin.02.00:Period.Start.Date WHERE [Period.Type.Code = 'NonResidentPartPeriod']

                    RP:^FTER43 = rvctc3.02.00:Elections.Commencement.Date
                */
        protected void VRATOFTER408085(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.FTER43 != null && report.TrusteeCollection != null)
                response.AddRange(report.TrusteeCollection
                    .Where(f => f.Trustee_PeriodDetailsNonResidentPartPeriodRPCollection?.Find(r => r.FTER38 != null && r.FTER38.GetValueOrDefault() < report.FTER43.GetValueOrDefault()) != null)
                    .Select(f =>
                        new ProcessMessageDocument
                        {
                            Code = "CMN.ATO.FTER.408085",
                            Severity = ProcessMessageSeverity.Error,
                            Description = "Trustee non-resident period must not be before the election commencement time",
                            LongDescription = "The start time during the relevant period at which the trustee was a non-resident for tax purposes must not be earlier than start of the election commencement time",
                            Location = $"/xbrli:xbrl/tns:Trustee/tns:PeriodDetails{OccurrenceIndex(1 + f.Trustee_PeriodDetailsNonResidentPartPeriodRPCollection.FindIndex(r => r.FTER38 != null && r.FTER38.GetValueOrDefault() < report.FTER43.GetValueOrDefault()))}/tns:Period.Start.Date",
                            Parameters = new ProcessMessageParameters
                            {
                                new ProcessMessageParameter {Name = "RuleIdentifier", Value = "VR.ATO.FTER.408085"},
                                new ProcessMessageParameter {Name = "FTER43", Value = GetValueOrEmptyDate(report.FTER43)}
                            }
                        }
                    ));
        }
        #endregion // VR.ATO.FTER.408085
        #region VR.ATO.FTER.408086
        /*  VR.ATO.FTER.408086
                Trustee non-resident period must not be before the start of income year specified

                Legacy Rule Format:
                    IF [FTER40] <> NULL AND [FTER41] = NULL AND [FTER43] = NULL AND (ANY OCCURRENCE OF [FTER38] <> NULL) AND ([FTER38] < ([FTER40] - 1)&"-07-01")
                       RETURN VALIDATION MESSAGE
                    ENDIF
                    [FTER40] = FTER:RP:Elections.FamilyTrustElectionStatus.Year
                    [FTER41] = FTER:RP:Period.Start.Date WHERE (TUPLE ELEMENT EXPLICIT pyin.xx.xx:Period.Type.Code = "SubstitutedAccountingPeriod") IN TUPLE(perioddetails1.xx.xx:PeriodDetails)
                    [FTER43] = FTER:RP:Elections.Commencement.Date
                    [FTER38] = FTER:RP:Period.Start.Date WHERE (TUPLE ELEMENT EXPLICIT pyin.xx.xx:Period.Type.Code = "NonResidentPartPeriod") IN TUPLE(perioddetails1.xx.xx:PeriodDetails) IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee)

                Technical Business Rule Format:
                    ^FTER40 <> NULL AND ^FTER41 = NULL AND ^FTER43 = NULL AND (AnyOccurrence(^FTER38, ^FTER38 <> NULL AND ^FTER38 < AsDate(Concat(^FTER40 - 1, '-07-01'))))

                Data Elements:

                    RP:^FTER38 = pyin.02.00:Period.Start.Date WHERE [Period.Type.Code = 'NonResidentPartPeriod']

                    RP:^FTER40 = rvctc3.02.00:Elections.FamilyTrustElectionStatus.Year

                    RP:^FTER41 = pyin.02.00:Period.Start.Date WHERE [Period.Type.Code = 'SubstitutedAccountingPeriod']

                    RP:^FTER43 = rvctc3.02.00:Elections.Commencement.Date
                */
        protected void VRATOFTER408086(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.FTER40 != null && report.FTER41 == null && report.FTER43 == null && report.TrusteeCollection != null)
                response.AddRange(report.TrusteeCollection
                    .Where(f => f.Trustee_PeriodDetailsNonResidentPartPeriodRPCollection?.Find(r => r.FTER38 != null && r.FTER38.GetValueOrDefault() < AsDate(string.Concat(report.FTER40.GetValueOrDefault() - 1, "-07-01"))) != null)
                    .Select(f =>
                        new ProcessMessageDocument
                        {
                            Code = "CMN.ATO.FTER.408086",
                            Severity = ProcessMessageSeverity.Error,
                            Description = "Trustee non-resident period must not be before the start of income year specified",
                            LongDescription = "The start time during the relevant period at which the trustee was a non-resident for tax purposes must not be earlier than 1 July of the income year specified",
                            Location = $"/xbrli:xbrl/tns:Trustee/tns:PeriodDetails{OccurrenceIndex(1 + f.Trustee_PeriodDetailsNonResidentPartPeriodRPCollection.FindIndex(r => r.FTER38 != null && r.FTER38.GetValueOrDefault() < report.FTER41.GetValueOrDefault()))}/tns:Period.Start.Date",
                            Parameters = new ProcessMessageParameters
                            {
                                new ProcessMessageParameter {Name = "RuleIdentifier", Value = "VR.ATO.FTER.408086"},
                                new ProcessMessageParameter { Name = "FTER40", Value = report.FTER40.GetValueOrDefault().ToString() },
                                new ProcessMessageParameter{ Name = "FTER41", Value = GetValueOrEmptyDate(report.FTER41) },
                                new ProcessMessageParameter {Name = "FTER43", Value = GetValueOrEmptyDate(report.FTER43)}
                            }
                        }
                    ));
        }

        #endregion // VR.ATO.FTER.408086
        #region VR.ATO.FTER.408164
        /*  VR.ATO.FTER.408164
                One Declaration Statement Type Code must be 'TrueAndCorrect'

                Legacy Rule Format:
                    IF COUNT(INT) = 0 AND 
                        COUNT (pyin.xx.xx:Declaration.StatementType.Code ="TrueAndCorrect" IN TUPLE(fter.0002.lodge.req.xx.xx:DeclarationOfTrustee\declaration2.xx.xx:Declaration)) <> 1
                      RETURN VALIDATION MESSAGE
                    ENDIF

                Technical Business Rule Format:
                    Count(^Context) = 0 AND Count(^FTER99 = 'TrueAndCorrect') <> 1

                Data Elements:

                    RP:^FTER99 = pyin.02.03:Declaration.StatementType.Code WHERE [Declaration.StatementType.Code = 'TrueAndCorrect']

                    RP:^Context = INT
                */
        protected void VRATOFTER408164(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.INTCount == 0 && report.DeclarationOfTrusteeCollection != null)
            {
                var collectionsRpTrueAndCorrect = report.DeclarationOfTrusteeCollection.Where(f => f.DeclarationOfTrustee_DeclarationCollection_TrueAndCorrect_RPCount == 1);
                if (collectionsRpTrueAndCorrect.Count() > 1)
                {
                    response.AddRange(collectionsRpTrueAndCorrect
                        .Skip(1)
                        .Select(f =>
                            new ProcessMessageDocument
                            {
                                Code = "CMN.ATO.FTER.408164",
                                Severity = ProcessMessageSeverity.Error,
                                Description = "Trustee declaration statement type code must have one 'TrueAndCorrect'",
                                LongDescription = "The FTER form must have one declaration that has type 'TrueAndCorrect'",
                                Location = "/xbrli:xbrl/tns:DeclarationOfTrustee/tns:Declaration/tns:Declaration.StatementType.Code",
                                Parameters = new ProcessMessageParameters { new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408164" } }
                            }
                        ));
                }
                else if (collectionsRpTrueAndCorrect.Count() < 1)
                {
                    response.Add(new ProcessMessageDocument
                    {
                        Code = "CMN.ATO.FTER.408164",
                        Severity = ProcessMessageSeverity.Error,
                        Description = "Trustee declaration statement type code must have one 'TrueAndCorrect'",
                        LongDescription = "The FTER form must have one declaration that has type 'TrueAndCorrect'",
                        Location = "/xbrli:xbrl/tns:DeclarationOfTrustee/tns:Declaration/tns:Declaration.StatementType.Code",
                        Parameters = new ProcessMessageParameters { new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408164" } }
                    });
                }
            }
        }
        #endregion // VR.ATO.FTER.408164
        #region VR.ATO.FTER.408166
        /*  VR.ATO.FTER.408166

                Legacy Rule Format:
                    WHERE IN TUPLE(Declaration2.xx.xx:Declaration)
                    IF pyin.xx.xx:Declaration.StatementAccepted.Indicator <> TRUE OR Declaration.SignatoryIdentifier.Text = NULLORBLANK
                       RETURN VALIDATION MESSAGE
                    ENDIF

                Technical Business Rule Format:
                    ^FTER169 <> TRUE OR ^FTER101 = BLANK OR ^FTER168 = NULL

                Data Elements:

                    INT:^FTER169 = pyin.02.28:Declaration.StatementAccepted.Indicator WHERE [Declaration.StatementType.Code = 'TrueAndCorrect']

                    INT:^FTER101 = pyin.02.00:Declaration.SignatoryIdentifier.Text WHERE [Declaration.StatementType.Code = 'TrueAndCorrect']
                */
        protected void VRATOFTER408166(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.DeclarationOfTrusteeCollection != null)
                response.AddRange(report.DeclarationOfTrusteeCollection
                    .Where(f => f.DeclarationOfTrustee_DeclarationCollection_TrueAndCorrect_RPCount == 1 && (f.FTER169 != true || string.IsNullOrWhiteSpace(f.FTER100)))
                    .Select(f =>
                        new ProcessMessageDocument
                        {
                            Code = "CMN.ATO.FTER.408167",
                            Severity = ProcessMessageSeverity.Error,
                            Description = "The Declaration must be complete",
                            LongDescription = "Within the Declaration, Statement Accepted must be 'Yes' (true), and the Signatory Identifier and Signature Date must be present.",
                            Location = "/xbrli:xbrl/tns:DeclarationOfTrustee/tns:Declaration/tns:Declaration.StatementType.Code",
                            Parameters = new ProcessMessageParameters
                            {
                                new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408166" },
                                new ProcessMessageParameter { Name = "FTER169", Value = f.FTER169.GetValueOrDefault().ToString() },
                                new ProcessMessageParameter { Name = "FTER100", Value = f.FTER100 },
                                //new ProcessMessageParameter { Name = "FTER168", Value = f.FTER168.GetValueOrDefault().ToString() }
                            }
                        }
                    ));
        }

        /*  VR.ATO.FTER.408166

                Legacy Rule Format:
                    WHERE IN TUPLE(Declaration2.xx.xx:Declaration)
                    IF pyin.xx.xx:Declaration.StatementAccepted.Indicator <> TRUE OR Declaration.SignatoryIdentifier.Text = NULLORBLANK
                       RETURN VALIDATION MESSAGE
                    ENDIF

                Technical Business Rule Format:
                    ^FTER160 <> TRUE OR ^FTER161 = BLANK OR ^FTER162 = NULL

                Data Elements:

                    INT:^FTER160 = pyin.02.28:Declaration.StatementAccepted.Indicator WHERE [Declaration.StatementType.Code = 'TrueAndCorrect']

                    INT:^FTER163 = pyin.02.00:Declaration.SignatoryIdentifier.Text WHERE [Declaration.StatementType.Code = 'TrueAndCorrect']
                */
        protected void VRATOFTER408166_(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            bool assertion = report.DeclarationCollection_TrueAndCorrect_INTExists && (report.FTER160 != true || string.IsNullOrWhiteSpace(report.FTER163));

            if (assertion)
            {
                var processMessage = new ProcessMessageDocument
                {
                    Code = "CMN.ATO.FTER.408167",
                    Severity = ProcessMessageSeverity.Error,
                    Description = "The Declaration must be complete",
                    LongDescription = "Within the Declaration, Statement Accepted must be 'Yes' (true), and the Signatory Identifier and Signature Date must be present.",
                    Location = $"/xbrli:xbrl/tns:Declaration{OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_INTOccurrenceIndex)}/tns:Declaration.StatementType.Code",
                    Parameters = new ProcessMessageParameters
                    {
                        new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408166" },
                        new ProcessMessageParameter { Name = "FTER160", Value = report.FTER160.GetValueOrDefault().ToString() },                    
                        new ProcessMessageParameter { Name = "FTER163", Value = report.FTER163 }
                    },
                };

                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.FTER.408166
        #region VR.ATO.FTER.408167
        /*  VR.ATO.FTER.408167

                Legacy Rule Format:
                    IF ([FTER159] <> NULLORBLANK) AND ([FTER159] <> "TrueAndCorrect" )
                       RETURN VALIDATION MESSAGE
                    ENDIF
                    
                    [FTER159] = FTER:INT:pyin.02.03:Declaration.StatementType.Code WHERE (pyin.02.03:Declaration.StatementType.Code = "TrueAndCorrect") IN TUPLE(fter.0002.lodge.req.02.00:Declaration)

                Technical Business Rule Format:
                    (^FTER159 <> BLANK) AND (^FTER159 <> 'TrueAndCorrect')

                Data Elements:
            
                    INT:^FTER159 = tns:Declaration.StatementType.Code WHERE [Declaration.StatementType.Code='TrueAndCorrect']
                    */
        protected void VRATOFTER408167(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            bool assertion = !string.IsNullOrWhiteSpace(report.FTER159) && !report.FTER159.Equals("TrueAndCorrect");

            if (assertion)
            {
                var processMessage = new ProcessMessageDocument
                {
                    Code = "CMN.ATO.GEN.430250",
                    Severity = ProcessMessageSeverity.Error,
                    Description = "The Declaration must be complete",
                    LongDescription = "Within the Declaration, the Statement Type must be 'TrueAndCorrect', Statement Accepted must be 'Yes' (true), and the Signatory Identifier and Signature Date must be present.",
                    Location = $"/xbrli:xbrl/tns:Declaration{OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_INTOccurrenceIndex)}/tns:Declaration.StatementType.Code",
                    Parameters = new ProcessMessageParameters
                    {
                        new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408167" },
                        new ProcessMessageParameter { Name = "FTER159", Value = report.FTER159 }
                    }
                };

                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.FTER.408167
        #region VR.ATO.FTER.408031
        /*  VR.ATO.FTER.408031
                Variation must not include central management or trustee details

                Legacy Rule Format:
                    IF [FTER95] = "V" AND ( [FTER16] <> NULL OR [FTER17] <> NULL OR COUNT (TUPLE(xbrli\perioddetails1.xx.xx.PeriodDetails
                    WHERE TUPLE ELEMENT EXPLICIT pyin.xx.xx:Period.Type.Code = "ControlOutsideAustraliaPartPeriod")) > 0 OR COUNT (TUPLE(fter.0002.lodge.req.xx.xx:Trustee)) > 0   )
                       RETURN VALIDATION MESSAGE
                    ENDIF
                    
                    [FTER16] = FTER:RP:pyde.02.00:Residency.CentralManagementAndControlOutsideAustralia.Indicator
                    [FTER17] = FTER:RP:pyde.02.00:Residency.CentralManagementAndControlOutsideAustraliaFullPeriod.Indicator
                    [FTER95] = FTER:RP:rvctc3.02.03:Elections.FamilyTrustElectionRevocation.Code

                Technical Business Rule Format:
                    ^FTER95 = 'V' AND (^FTER16 <> NULL OR ^FTER17 <> NULL OR Count(^FTER88) > 0 OR Count(^Trustee) > 0)

                Data Elements:
            
                    RP:^FTER16 = tns:Residency.CentralManagementAndControlOutsideAustralia.Indicator
            
                    RP:^FTER17 = tns:Residency.CentralManagementAndControlOutsideAustraliaFullPeriod.Indicator
            
                    RP:^FTER95 = tns:Elections.FamilyTrustElectionRevocation.Code
            
                    ^Trustee = tns:Trustee
            
                    RP:^FTER88 = tns:Period.Type.Code WHERE [Period.Type.Code='ControlOutsideAustraliaPartPeriod']
                */
        protected void VRATOFTER408031(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if(report.FTER95 == "V")
            {
                int countPeriodDetailsControlOutsideAustraliaPartPeriod = report.PeriodDetailsControlOutsideAustraliaPartPeriodRPCollection?.Count(f => f.FTER88.Equals("ControlOutsideAustraliaPartPeriod")) ?? 0;

                bool assertion = report.FTER16 != null || report.FTER17 != null || countPeriodDetailsControlOutsideAustraliaPartPeriod > 0 || report.TrusteeCollectionCount > 0;

                if (assertion)
                {
                    var processMessage = new ProcessMessageDocument
                    {
                        Code = "CMN.ATO.FTER.408031",
                        Severity = ProcessMessageSeverity.Error,
                        Description = "Variation must not include central mgmnt or trustee details",
                        LongDescription = "If the form is used to advise of a variation, details regarding central management and control and of the trustees must not be completed",
                        Location = "/xbrli:xbrl/tns:Residency.CentralManagementAndControlOutsideAustralia.Indicator",
                        Parameters = new ProcessMessageParameters
                        {
                            new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408031" },
                            new ProcessMessageParameter { Name = "FTER95", Value = report.FTER95 },
                            new ProcessMessageParameter { Name = "FTER16", Value = GetValueOrEmpty(report.FTER16) },
                            new ProcessMessageParameter { Name = "FTER17", Value = GetValueOrEmpty(report.FTER17) }

                        }
                    };

                    response.Add(processMessage);
                }
            }
        }
        #endregion // VR.ATO.FTER.408031
        #region VR.ATO.FTER.408051
        /*  VR.ATO.FTER.408051
                Dates central management and control of trust was outside Australia must be present

                Legacy Rule Format:
                    IF [FTER16] = TRUE AND [FTER17] <> TRUE AND COUNT (TUPLE(xbrli\perioddetails1.xx.xx.PeriodDetails WHERE TUPLE ELEMENT EXPLICIT pyin.xx.xx:Period.Type.Code = "ControlOutsideAustraliaPartPeriod")) = 0
                       RETURN VALIDATION MESSAGE
                    ENDIF
                    
                    [FTER16] = FTER:RP:pyde.02.00:Residency.CentralManagementAndControlOutsideAustralia.Indicator
                    [FTER17] = FTER:RP:pyde.02.00:Residency.CentralManagementAndControlOutsideAustraliaFullPeriod.Indicator

                Technical Business Rule Format:
                    ^FTER16 = TRUE AND ^FTER17 <> TRUE AND Count(^FTER88) = 0

                Data Elements:
            
                    RP:^FTER16 = tns:Residency.CentralManagementAndControlOutsideAustralia.Indicator
            
                    RP:^FTER17 = tns:Residency.CentralManagementAndControlOutsideAustraliaFullPeriod.Indicator
            
                    RP:^FTER88 = tns:Period.Type.Code WHERE [Period.Type.Code='ControlOutsideAustraliaPartPeriod']
                */
        protected void VRATOFTER408051(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.FTER16.GetValueOrDefault())
            {
                int countPeriodDetailsControlOutsideAustraliaPartPeriod = report.PeriodDetailsControlOutsideAustraliaPartPeriodRPCollection?.Count(f => f.FTER88.Equals("ControlOutsideAustraliaPartPeriod")) ?? 0;

                bool assertion =  !report.FTER17.GetValueOrDefault() && countPeriodDetailsControlOutsideAustraliaPartPeriod == 0;

                if (assertion)
                {
                    var processMessage = new ProcessMessageDocument
                    {
                        Code = "CMN.ATO.FTER.408051",
                        Severity = ProcessMessageSeverity.Error,
                        Description = "Dates central management and control of trust was outside Australia must be present",
                        LongDescription = "If the answer to the central management and control of the trust outside Australia question is 'Yes' and central management and control of the trust was outside Australia for Full Period indicator is not 'yes' (true), then the dates that the management and control of the trust were outside Australia must be present",
                        Location = "/xbrli:xbrl/tns:Residency.CentralManagementAndControlOutsideAustralia.Indicator",
                        Parameters = new ProcessMessageParameters
                        {
                            new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408051" },
                            new ProcessMessageParameter{ Name = "FTER16", Value = GetValueOrEmpty(report.FTER16) },
                            new ProcessMessageParameter { Name = "FTER17", Value = GetValueOrEmpty(report.FTER17) }
                        }
                    };

                    response.Add(processMessage);
                }
            }
        }
        #endregion // VR.ATO.FTER.408051
        #region VR.ATO.FTER.408052

        /*  VR.ATO.FTER.408052
                Time central management and control of trust was outside Australia must NOT be present

                Legacy Rule Format:
                    IF [FTER16] = FALSE AND ([FTER17] <> NULL OR COUNT (TUPLE(xbrli\perioddetails1.xx.xx.PeriodDetails WHERE TUPLE ELEMENT EXPLICIT pyin.xx.xx:Period.Type.Code = "ControlOutsideAustraliaPartPeriod"))) > 0
                       RETURN VALIDATION MESSAGE
                    ENDIF
                    
                    [FTER16] = FTER:RP:pyde.02.00:Residency.CentralManagementAndControlOutsideAustralia.Indicator
                    [FTER17] = FTER:RP:pyde.02.00:Residency.CentralManagementAndControlOutsideAustraliaFullPeriod.Indicator

                Technical Business Rule Format:
                    ^FTER16 = FALSE AND (^FTER17 <> NULL OR Count(^FTER88) > 0)

                Data Elements:
            
                    RP:^FTER16 = tns:Residency.CentralManagementAndControlOutsideAustralia.Indicator
            
                    RP:^FTER17 = tns:Residency.CentralManagementAndControlOutsideAustraliaFullPeriod.Indicator
            
                    RP:^FTER88 = tns:Period.Type.Code WHERE [Period.Type.Code='ControlOutsideAustraliaPartPeriod']
                */
        protected void VRATOFTER408052(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.FTER16.HasValue && !report.FTER16.GetValueOrDefault())
            {
                int countPeriodDetailsControlOutsideAustraliaPartPeriod = report.PeriodDetailsControlOutsideAustraliaPartPeriodRPCollection?.Count(f => f.FTER88.Equals("ControlOutsideAustraliaPartPeriod")) ?? 0;

                bool assertion = report.FTER17 != null || countPeriodDetailsControlOutsideAustraliaPartPeriod > 0;

                if (assertion)
                {
                    var processMessage = new ProcessMessageDocument
                    {
                        Code = "CMN.ATO.FTER.408052",
                        Severity = ProcessMessageSeverity.Error,
                        Description = "Time central management and control of trust was outside Australia must NOT be present",
                        LongDescription = "If the answer to the central management and control of the trust was outside Australia is 'No' (false), then the Full Period indicator and dates must be left blank.",
                        Location = "/xbrli:xbrl/tns:Residency.CentralManagementAndControlOutsideAustralia.Indicator",
                        Parameters = new ProcessMessageParameters
                        {
                            new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408052" },
                            new ProcessMessageParameter { Name = "FTER16", Value = GetValueOrEmpty(report.FTER16) },
                            new ProcessMessageParameter { Name = "FTER17", Value = GetValueOrEmpty(report.FTER17) }
                        },
                    };

                    response.Add(processMessage);
                }
            }
        }
        #endregion // VR.ATO.FTER.408052
        #region VR.ATO.FTER.408055
        /*  VR.ATO.FTER.408055
                Central management and control - full period indicator and dates must not both be present

                Legacy Rule Format:
                    IF [FTER17] = TRUE AND COUNT (TUPLE(xbrli\perioddetails1.xx.xx.PeriodDetails WHERE TUPLE ELEMENT EXPLICIT pyin.xx.xx:Period.Type.Code = "ControlOutsideAustraliaPartPeriod")) > 0
                       RETURN VALIDATION MESSAGE
                    ENDIF
                    
                    [FTER17] = FTER:RP:pyde.02.00:Residency.CentralManagementAndControlOutsideAustraliaFullPeriod.Indicator

                Technical Business Rule Format:
                    ^FTER17 = TRUE AND Count(^FTER88) > 0

                Data Elements:
            
                    RP:^FTER17 = tns:Residency.CentralManagementAndControlOutsideAustraliaFullPeriod.Indicator
            
                    RP:^FTER88 = tns:Period.Type.Code WHERE [Period.Type.Code='ControlOutsideAustraliaPartPeriod']
                */
        protected void VRATOFTER408055(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.FTER17.GetValueOrDefault())
            {
                int countPeriodDetailsControlOutsideAustraliaPartPeriod = report.PeriodDetailsControlOutsideAustraliaPartPeriodRPCollection?.Count(f => f.FTER88.Equals("ControlOutsideAustraliaPartPeriod")) ?? 0;

                bool assertion = countPeriodDetailsControlOutsideAustraliaPartPeriod > 0;

                if (assertion)
                {
                    var processMessage = new ProcessMessageDocument
                    {
                        Code = "CMN.ATO.FTER.408055",
                        Severity = ProcessMessageSeverity.Error,
                        Description = "Central management and control - full period indicator and dates must not both be present",
                        LongDescription = "If Full period indicator for Central management and control outside Australia is 'yes' (true), then specific dates must not be present",
                        Location = "/xbrli:xbrl/tns:Residency.CentralManagementAndControlOutsideAustraliaFullPeriod.Indicator",
                        Parameters = new ProcessMessageParameters
                        {
                            new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408055" },
                            new ProcessMessageParameter { Name = "FTER17", Value = GetValueOrEmpty(report.FTER17) }
                        },
                    };
                    response.Add(processMessage);
                }
            }
        }
        #endregion // VR.ATO.FTER.408055
        #region VR.ATO.FTER.408081
        /*  VR.ATO.FTER.408081
                Trustee non-resident Period Type Code must be 'NonResidentPartPeriod'

                Legacy Rule Format:
                    WHERE IN TUPLE(fter.0002.lodge.req.xx.xx:Trustee\perioddetails1.xx.xx.PeriodDetails)
                    IF pyid.xx.xx:Period.Type.Code <> "NonResidentPartPeriod"
                       RETURN VALIDATION MESSAGE
                    ENDIF

                Technical Business Rule Format:
                    ^FTER90 <> 'NonResidentPartPeriod'

                Data Elements:
            
                    RP:^FTER90 = tns:Period.Type.Code WHERE [Period.Type.Code='NonResidentPartPeriod']
                */
        protected void VRATOFTER408081(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            if (report.TrusteeCollection != null)
                response.AddRange(report.TrusteeCollection
                    .Where(f => f.Trustee_PeriodDetailsCollectionCount != Count(f.Trustee_PeriodDetailsNonResidentPartPeriodRPCollection))
                    .Select(f =>
                        new ProcessMessageDocument
                        {
                            Code = "CMN.ATO.FTER.408081",
                            Severity = ProcessMessageSeverity.Error,
                            Description = "Trustee non-resident Period Type Code must be 'NonResidentPartPeriod'",
                            Location = "/xbrli:xbrl/tns:Trustee/tns:PeriodDetails/tns:Period.Type.Code",
                            Parameters = new ProcessMessageParameters { new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408081" } }
                        }
                    ));
        }
        #endregion // VR.ATO.FTER.408081
        #region VR.ATO.FTER.408056
        /*  VR.ATO.FTER.408056
                    Number of Central Management and Control Period details supplied cannot exceed 12

                Legacy Rule Format:
                    IF COUNT (TUPLE(xbrli\perioddetails1.xx.xx.PeriodDetails WHERE TUPLE ELEMENT EXPLICIT pyin.xx.xx:Period.Type.Code = "ControlOutsideAustraliaPartPeriod")) > 12
                       RETURN VALIDATION  MESSAGE
                    ENDIF

                Technical Business Rule Format:
                    Count(^FTER88) > 12

                Data Elements:
            
                    RP:^FTER88 = tns:Period.Type.Code WHERE [Period.Type.Code='ControlOutsideAustraliaPartPeriod']
                    */
        protected void VRATOFTER408056(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            bool assertion = Count(report.PeriodDetailsControlOutsideAustraliaPartPeriodRPCollection) > 12;

            if (assertion)
            {
                var processMessage = new ProcessMessageDocument
                {
                    Code = "CMN.ATO.FTER.408056",
                    Severity = ProcessMessageSeverity.Error,
                    Description = "Number of separate Periods of central management and control must not exceed 12",
                    Location = "/xbrli:xbrl/tns:PeriodDetails/tns:Period.Start.Date",
                    Parameters = new ProcessMessageParameters { new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408056" } }
                };

                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.FTER.408056
        #region VR.ATO.GEN.408058
        /*  VR.ATO.GEN.408058

                Legacy Rule Format:
                    WHERE IN TUPLE (xbrli\perioddetails1.xx.xx.PeriodDetails)
                    IF pyin.xx.xx:Period.Type.Code <> NULLORBLANK AND pyin.xx.xx:Period.Type.Code <> SET("ControlOutsideAustraliaPartPeriod", "SubstitutedAccountingPeriod")
                       RETURN VALIDATION MESSAGE
                    ENDIF

            Technical Business Rule Format:
                    ((^FTER88 <> BLANK) AND (^FTER88 <> 'ControlOutsideAustraliaPartPeriod' AND ^FTER88 <> 'SubstitutedAccountingPeriod')) OR ((^FTER89 <> BLANK) AND (^FTER89 <> 'ControlOutsideAustraliaPartPeriod' AND ^FTER89 <> 'SubstitutedAccountingPeriod'))

        Data Elements:
            
                    RP:^FTER88 = tns:Period.Type.Code WHERE [Period.Type.Code='ControlOutsideAustraliaPartPeriod']
            
                    RP:^FTER89 = tns:Period.Type.Code WHERE [Period.Type.Code='SubstitutedAccountingPeriod']
                    */
        protected void VRATOGEN408058(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {
            bool assertion = report.PeriodDetailsCollectionCount != Count(report.PeriodDetailsControlOutsideAustraliaPartPeriodRPCollection) + Count(report.PeriodDetailsCollection_SubstitutedAccountingPeriod_RPCount);

            if (assertion)
            {
                var processMessage = new ProcessMessageDocument
                {
                    Code = "CMN.ATO.GEN.408058",
                    Severity = ProcessMessageSeverity.Error,
                    Description = "Period Type Code must be 'ControlOutsideAustraliaPartPeriod' or 'SubstitutedAccountingPeriod'",
                    LongDescription = "For each central management and control outside Australia period and the Substituted Account Period, the Period Type Code must be 'ControlOutsideAustraliaPartPeriod' or 'SubstitutedAccountingPeriod', respectively",
                    Location = "/xbrli:xbrl/tns:PeriodDetails/tns:Period.Type.Code",
                    Parameters = new ProcessMessageParameters { new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.GEN.408058" } }
                };

                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.GEN.408058
        #region VR.ATO.FTER.408169

        /*  VR.ATO.FTER.408169
                    Trust Address Details Usage Code must be 'POS' and Currency Code must be 'C' or 'P'

                Legacy Rule Format:
                    WHERE IN TUPLE (xbrli\address2.xx.xx:AddressDetails)
                    IF (AddressDetails.Usage.Code <> "POS") OR (AddressDetails.Currency.Code <> "C" AND AddressDetails.Currency.Code <> "P")
                       RETURN VALIDATION MESSAGE
                    ENDIF

            Technical Business Rule Format:
                    (AddressDetails.Usage.Code <> 'POS') OR (AddressDetails.Currency.Code <> 'C' AND AddressDetails.Currency.Code <> 'P')

        Data Elements:
        */
        protected void VRATOFTER408169(List<ProcessMessageDocument> response, FTER2018 report, bool isStandalone)
        {

            bool assertion =( report.AddressDetailsCollection_POS_C_RPCount + report.AddressDetailsCollection_POS_P_RPCount
                             != report.AddressDetailsCollectionCount);

            if (assertion)
            {
                int locationIndex = 0;
                for (int i = 1; i <= report.AddressDetailsCollectionCount; i++)
                {
                    if (report.AddressDetailsCollection_POS_P_RPOccurrenceIndex !=i &&
                       report.AddressDetailsCollection_POS_C_RPOccurrenceIndex !=i)
                    {
                        locationIndex = i;
                        break;
                    }
                }

               var processMessage = new ProcessMessageDocument
                    {
                        Code = "CMN.ATO.FTER.408169",
                        Severity = ProcessMessageSeverity.Error,
                        Description = "Trust Address Details Usage Code must be 'POS' and Currency Code must be 'C' or 'P'",
                        LongDescription = "Trust Address Details Usage Code must be 'POS' (Postal) and Currency Code must be 'C' (Current) or 'P' (Previous)",
                        Location = (locationIndex<=1) ? "/xbrli:xbrl/tns:AddressDetails/tns:AddressDetails.Usage.Code": "/xbrli:xbrl/tns:AddressDetails["+ locationIndex +"]/tns:AddressDetails.Usage.Code",
                       Parameters = new ProcessMessageParameters { new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.FTER.408169" } }
                    };

                    response.Add(processMessage);
                }
           
        }


       
        #endregion // VR.ATO.FTER.408169

        #endregion Manual Rules - Rules marked for manual coding
    }
}